import java.io.*;
import java.net.*;
import java.util.*;

public class BNet
{
    protected int nodes;        // number of nodes
    protected boolean [][] arc; // directed edges
    protected String [] name;   // names of nodes
    protected int [] values;    // number of values for each node
    protected double [][][] cpt;  // conditional probability tables
    protected Random rand;

    public BNet()
    { 
	rand = new Random();
	nodes = 0;
	arc = null;
	name = null;
	values = null;
	cpt = null;
    }

    public BNet(int nodes)
    {
	rand = new Random();
	this.nodes = nodes;	
	arc = new boolean[nodes][nodes];
	name = new String[nodes];
	values = new int[nodes];
	cpt = new double[nodes][][];
	for (int i = 0; i < nodes; i++)
	{
	    name[i] = new Integer(i).toString();
	    values[i] = 2;
	    for (int j = 0; j < nodes; j++)
		arc[i][j] = false;
	}
    }

    public BNet(BNet bnet)
    {
	rand = new Random();
	this.nodes = bnet.nodes;
	arc = new boolean[nodes][nodes];
	name = new String[nodes];
	values = new int[nodes];
	cpt = new double[nodes][][];
	for (int i = 0; i < nodes; i++)
	{
	    name[i] = bnet.getName(i);
	    values[i] = bnet.values[i];
	    for (int j = 0; j < nodes; j++)
		arc[i][j] = bnet.isArc(i, j);
	}
    }

    public int getValues(int i)
    {
	return values[i];
    }

    public int getParentConfigurations(int i)
    {
	int pacs = 1;

	for (int j = 0; j < nodes; j++)
	    if (isArc(j, i))
		pacs *= values[j];

	return pacs;
    }

    protected static String skipComments(BufferedReader dataReader)
    {
	String line;

	try {
	    while ((line = dataReader.readLine()) != null &&
		   line.length() > 0 &&
		   line.charAt(0) == '#');
	}
	catch (Exception e)
	{
	    System.out.println(e);
	    return null;
	}
	    
	return line;
    }

    public static BNet readFromFile(String fname)
    {
        BufferedReader dataReader = null;
	BNet bnet;
	String line = null;
	int fs = -2, nextFs; // field separators

        try {
            dataReader = new BufferedReader(new FileReader(fname));

            line = skipComments(dataReader);

	    // number of nodes

	    int nodes = new Integer(line).intValue();

	    bnet = new BNet(nodes);

	    // structure

            for (int node = 0; node < nodes; node++)
	    {
                line = skipComments(dataReader);
		fs = line.lastIndexOf('"');
		
		// name

		if (fs != -1)
		    bnet.setName(node, line.substring(1, fs));
		else
		    fs = -2;

		nextFs = line.indexOf(':');

		// number of values

		bnet.values[node] = 
		    new Integer(line.substring(fs+2, nextFs)).intValue();

		// parents

		fs = nextFs+2;

		while (fs < line.length())
		{
		    nextFs = line.indexOf(' ', fs);
		    if (nextFs == -1)
			nextFs = line.length();

		    String substr = line.substring(fs, nextFs);
		    if (substr != null)
		    {
			int node2 =
			    new Integer(substr).intValue();
			bnet.addArc(node2, node);
		    }
		    else
			break;

		    fs = nextFs+1;
		    nextFs = line.indexOf(' ', fs);
		}
            }

	    // conditional probability tables

            for (int node = 0; node < nodes; node++)
	    {
		int pacs = bnet.getParentConfigurations(node);

		for (int pac = 0; pac < pacs; pac++)
		{
		    line = skipComments(dataReader);

		    // if none available exit
		    if (line == null)
		    {
			node = nodes;
			break;
		    }

		    if (pac == 0)
			bnet.cpt[node] = new double[pacs][bnet.values[node]];
		    fs = 0;

		    double sum = 0.0;
		    for (int val = 0; val < bnet.values[node]; val++)
		    {
			nextFs = line.indexOf(' ', fs);
			if (nextFs == -1)
			    nextFs = line.length();
			String substr = line.substring(fs, nextFs);
			bnet.cpt[node][pac][val] = 
			    new Double(substr).doubleValue();
			sum += bnet.cpt[node][pac][val];
			fs = nextFs+1;
		    }
		    if (sum != 1.0)
			for (int val = 0; val < bnet.values[node]; val++)
			    bnet.cpt[node][pac][val] /= sum;
		}
		line = skipComments(dataReader); // discard empty line
	    }
        }
        catch (IOException e)
        {
            System.out.println(e + "\n" + line + "\n" + fs);
	    return null;
        }

        return bnet;
    }

    public int size()
    {
	return nodes;
    }

    public boolean hasCPTs()
    {
	if (cpt[0] == null)
	    return false;
	else
	    return true;
    }

    public double [][] getCPT(int i)
    {
	return cpt[i];
    }

    public String getName(int i)
    {
	return name[i];
    }
    
    public void setName(int i, String name)
    {
	this.name[i] = name;
    }

    public int getIndegree(int i)
    {
	int indegree = 0;

	for (int j = 0; j < nodes; j++)
	    if (isArc(j, i))
		indegree++;

	return indegree;
    }

    public boolean isArc(int i, int j)
    {
	return arc[i][j];
    }
    
    public boolean isEdge(int i, int j)
    {
	if (isArc(i, j) || isArc(j, i))
	    return true;
	else
	    return false;
    }

    public boolean addArc(int i, int j)
    {
	arc[i][j] = true;
	return true;
    }

    public boolean deleteArc(int i, int j)
    {
	if (!isArc(i, j))
	    return false;

	arc[i][j] = false;

	return true;
    }

    public int[][] getContexts(int i)
    {
	int parents = 0, contexts = 1;
	Vector vParent = new Vector();
	int [] parent;

	for (int k = 0; k < nodes; k++)
	    if (isArc(k, i))
	    {
		parents++;
		vParent.add(new Integer(k));
		contexts *= getValues(k);
	    }

	int[][] vec = new int[contexts][parents];
	parent = new int[parents];
	for (int k = 0; k < parents; k++)
	{
	    parent[k] = ((Integer)vParent.elementAt(k)).intValue();
	    vec[0][k] = 0;
	}

	for (int context = 1; context < contexts; context++)
	{
	    for (int k = 0; k < parents; k++)
		vec[context][k] = vec[context-1][k];
	    int l = parents-1;
	    while (++vec[context][l] >= getValues(parent[l]))
	    {
		vec[context][l] = 0;
		l--;
	    }
	}

	return vec;
    }

    private void sampleValue(int [] vec, int i)
    {
	int context = 0;

	if (vec[i] != -1)
	    return;

	for (int j = 0; j < nodes; j++)
	{
	    if (j != i && isArc(j, i))
	    {
		if (vec[j] == -1)
		    sampleValue(vec, j);

		context = context * values[j] + vec[j];
	    }
	}

	double cut = rand.nextDouble();
	double sum = 0.0;

	vec[i] = values[i]-1;
	for (int j = 0; j < values[i]-1; j++)
	{
	    sum += cpt[i][context][j];
	    if (cut < sum)
	    {
		vec[i] = j;
		break;
	    }
	}
    }

    public String toHeaderString()
    {
	StringBuffer head = new StringBuffer();

	for (int i = 0; i < nodes; i++)
	{
	    if (i > 0)
		head.append("\t");
	    head.append("\"" + getName(i) + "\" " + values[i]);
	}
	
	return head.toString();
    }

    public int[] genDataVector()
    {
	int [] vec = new int[nodes];

	for (int i = 0; i < nodes; i++)
	    vec[i] = -1;

	for (int i = 0; i < nodes; i++)
	    sampleValue(vec, i);

	return vec;
    }
}
